import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.Stack;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;
import java.util.Optional;


//Class: ExpSolver
//Contiene tutti i metodi necessari per ottenere una funzione booleana in sintesi minima partendo dalla sua rappresentazione postfissa. Ogni metodo risulta essere una trasposizione di un determinato step dell'algoritmo di quine-mccluskey. Poichè non risulta necessario instanziare più di un oggetto della classe, la suddetta è stata pensata come un singleton.

public class ExpSolver 
{
	private static ExpSolver e=null;
	private ExpSolver() {}
	
	public ExpSolver getExpSolver() 
	{
		if(e==null)e=new ExpSolver();
		return e;
	}
// Function: insertValues
// Restituisce una stringa ottenuta sostituendo in exp i lettarali con i valori ad essi associati.	
	private static String insertValues(String exp,List<String> inputValues,List<Character> variables)
	{
		return     exp.chars()
				  .map(x->{return x>=97&&x<=122?getValue((char)x,inputValues,variables):x;})
				  .mapToObj(x->new String()+(char)x).collect(Collectors.joining());
	}	
// Function: getValue 
//	Restituisce il valore (carattere) associato alla variabile in input.
	private static char getValue(char x,List<String> inputValues,List<Character> variables)
	{
		return inputValues.get(variables.indexOf(x)).charAt(0);
	}
// Function: computeTruthTable 
// Restituisce una lista contenente in ogni riga un oggetto Term,
// che contiene il valore di verita' che la funzione descritta da exp assume in corrispondenza dei valori di inputValue.	
	static List<Term> computeTruthTable(List<List<String>> inputValues,List<Character> variables,String exp)
	{
		 List<Term> truthTab=new LinkedList<>();		
		 int index=0;
		 for(List<String> x :inputValues)truthTab.add(new Term(x,index++,computeExp(insertValues(exp,x,variables))));
		 return truthTab;
	}
//	Function: computeMinTermTab
//  Restituisce la truthTab privata dei Term in corrispondenza dei quali la funzione assume valore 0.	
	public static List<Term> computeMinTermTab(List<Term> truthTab)
	{
		return truthTab.stream()
						 .filter(x->{return x.getValue().equals("1")?true:false;})
						 .collect(Collectors.toCollection(ArrayList::new));
	}
// Function: hammingDist
// Restituisce la distanza di hamming che intercorre tra i termini l ed l1.	
	private static int hammingDist(List<String> l,List<String> l1)
	{
		return IntStream.rangeClosed(0,l.size()-1).mapToObj(x->{
			return !l.get(x).equals(l1.get(x))?true:false;	
		}).filter(x->{return x==true?true:false;}).collect(Collectors.toList()).size();
	}
	public static List<Integer> getMinTermsIndex(List<ImpTerm> l)
	{
		List<Integer> i=new ArrayList<>();
		for (int x=0;x<l.size();x++)if(l.get(x)!=null)i.add(x);
		return i;
	}
// Function: duplicateMinCover
// Duplica la tabella di copertura in ingresso.	

	private static List<List<ImpTerm>> duplicateMinCover(List<ImpTerm> minCover,int n)
	{
		return IntStream.range(0,n).mapToObj(x->
				{
					return minCover.stream().collect(Collectors.toList());
				}).collect(Collectors.toList());
	}

//Function: simplifyCouples
//Riempie fimpTab con tutti gli implicanti primi che riesce a trovare.
	
	public static List<ImpTerm> simplifyCouples(List<ImpTerm> impTable,List<ImpTerm> fimpTab,List<List<ImpTerm>> couples)
	{
		List<ImpTerm> tempTab=new ArrayList<>();
		for(int i=0;i<impTable.size();i++)
		{
			for(int k=i+1;k<impTable.size();k++)
			{
				if(checkCondition(impTable,i,k))
				{
					List <Integer> minIndex=Stream.concat(impTable.get(i).getIndex().stream(),impTable.get(k).getIndex().stream())
												  .collect(Collectors.toList());
					ImpTerm imp=new ImpTerm(simplifyTerm(impTable.get(i).getTerms(),impTable.get(k).getTerms()),minIndex);
					couples.add(Arrays.asList(impTable.get(i),impTable.get(k),imp));
					if(!tempTab.contains(imp))tempTab.add(imp);
					impTable.get(i).setFirst();
					impTable.get(k).setFirst();
				}		
			}
			if(impTable.get(i).isFirst()&&!fimpTab.contains(impTable.get(i)))
				fimpTab.add(impTable.get(i));
		}
		return tempTab;
	}
//Function: checkCondition
//Controlla se i due termini differisco di un solo valore positivo e se la distanza di hamming tra i due termini è uno. 

	private static boolean checkCondition(List<ImpTerm> impTable,int i,int k)
	{
		boolean a= (Math.abs(getCountTrue(impTable.get(i).getTerms())-getCountTrue(impTable.get(k).getTerms())))==1?true:false;
		boolean b= hammingDist(impTable.get(i).getTerms(),impTable.get(k).getTerms())==1?true:false;
		return a&&b;
	}

//Function: computeFimpTab
//Calcola la tabella degli implicanti primi 
	
	static List<ImpTerm> computeFimpTab(String exp,List<Term> minTab,List<ImpTerm> fimpTab,List<List<ImpTerm>> tempFimpTab,List<List<ImpTerm>> couples)
	{
		Stack<List<ImpTerm>> s=new Stack<>();
		List<ImpTerm> tempTab=minTab.stream().map(x->{return new ImpTerm(x);}).collect(Collectors.toList());
		s.push(tempTab);
		tempFimpTab.add(tempTab);
		while(true)
		{
			tempTab=simplifyCouples(tempTab,fimpTab,couples);
			tempFimpTab.add(tempTab);
			if(tempTab.isEmpty())break;
			tempTab=tempTab.stream().filter(x->!x.fullySimplify()).collect(Collectors.toList());
			if(!s.lastElement().equals(tempTab))s.push(tempTab);
			else break;
		}	
		return Stream.concat(fimpTab.stream(),s.pop().stream()).distinct().collect(Collectors.toList());
	}

//Function: computeFimpTab
//Restituisce la tabella di copertura che può essere vuota nel caso in cui tutti gli imlicanti primi essenziali siano stati 
	
	static List<List<ImpTerm>> computeMinCover(List<List<ImpTerm>> coverTab,List<ImpTerm> minCover,List<List<List<ImpTerm>>> tempCoverTab)
	{
		if(coverTab.isEmpty())return coverTab;
		tempCoverTab.add(coverTab);
		while(true)
		{
			boolean simplified=false;
			int fimpIndex[]=null;
			if(fimpIndex!=null||coverTab.size()==1)
			{
				simplified=true;
				if(coverTab.size()==1)
				{
					minCover.add(coverTab.get(0).stream().filter(x->x!=null).collect(Collectors.toList()).get(0));
					coverTab=new ArrayList<>();
				}
				else 
				{
					minCover.add(coverTab.get(fimpIndex[0]).get(fimpIndex[1]));
					coverTab=removeRowAndCulums(fimpIndex[0],Arrays.asList(fimpIndex[1]),coverTab);
				}
			}			
			else 
			{
				List<List<ImpTerm>> t=simplifyCoverTab(coverTab);
				if(t!=null)
				{
					coverTab=t;
					simplified=true;
				}
			}
			tempCoverTab.add(coverTab);
			if(!simplified||coverTab.size()==0)return coverTab;
		}
	}

//Function: simplifyCoverTab
//Semplifica la tabella di copertura controllando un'eventuale dominanza di riga o di colonna.
	
	private static List<List<ImpTerm>> simplifyCoverTab(List<List<ImpTerm>> coverTab)
	{
			if (coverTab.isEmpty())return null;
			Stack<List<ImpTerm>> s=new Stack<>();
			List<List<ImpTerm>> temp=simplifyRows(coverTab,getIndexOfObj(coverTab,s),s);
			if(temp!=null)return temp;
			s=new Stack<>();
			temp=simplifyCulums(coverTab,getIndexOfCulums(getReverseTab(coverTab),s),s);
			if(temp!=null)return temp;
			return null;
	}

//Function: getCountOfObj
//Ritorna il numero degli elementi non nulli,per ogni riga, nella tabella.
	
	private static List<Integer> getIndexOfObj(List<List<ImpTerm>> coverTab,Stack<List<ImpTerm>> s)
	{

		List<Integer> t=new ArrayList<>();
		for (int i=0;i<coverTab.size();i++ )
		{
			List<Integer> t1=getIndex(coverTab.get(i));
			if(t1.size()>t.size())
			{
				if(!s.isEmpty())s.pop();
				s.push(coverTab.get(i));
				t=t1;
			}
		} 
		return t;
	}
	private static List<Integer> getIndexOfCulums(List<List<ImpTerm>> coverTab,Stack<List<ImpTerm>> s)
	{
		List<Integer> t=new ArrayList<>();
		for (int i=0;i<coverTab.size();i++)
		{
			List<Integer> t1=getIndex(coverTab.get(i));
			if(t1.size()<t.size()||t.size()==0)
			{
				if(!s.isEmpty())s.pop();
				s.push(coverTab.get(i));
				t=t1;
			}
		} 
		return t;
	}
	protected static List<Integer> getIndex(List<ImpTerm> l)
	{
		List<Integer> a=new ArrayList<>();
		for (int k=0;k<l.size();k++)if(l.get(k)!=null)a.add(k);
		return a;
	}

//Function: simplifyRows
//Restituisce la tabella privata della riga.

	private static List<List<ImpTerm>> simplifyRows(List<List<ImpTerm>> coverTab,List<Integer> index,Stack<List<ImpTerm>> s)
	{
		List<List<ImpTerm>> temp=coverTab.stream().filter(x->
		{
			return !checkIfRowIsDominated(index,getIndex(x))||x==s.lastElement()?true:false;
		}).collect(Collectors.toList());
		return coverTab.equals(temp)?null:temp;
	} 

	protected static boolean checkIfRowIsDominated(List<Integer> index,List<Integer>index1)
	{
		int e=0;
		for (Integer x : index1)if(index.contains(x))e++;
		return e==index1.size()?true:false;
	}

//Function: simplifyRows
//Restituisce la tabella privata della colonna.

	private static List<List<ImpTerm>> simplifyCulums(List<List<ImpTerm>> coverTab,List<Integer> index,Stack<List<ImpTerm>> s)
	{
		List<List<ImpTerm>> temp= getReverseTab(getReverseTab(coverTab).stream().filter(x->
		{
			return !checkIfCulumsIsDominated(index,getIndex(x))||x==s.lastElement()?true:false;
		}).collect(Collectors.toList()));
		return coverTab.equals(temp)?null:temp;
	}

	protected static boolean checkIfCulumsIsDominated(List<Integer> index,List<Integer>index1)
	{
		int e=0;
		for (Integer x : index)if(index1.contains(x))e++;
		return e==index.size()&&e<index1.size()?true:false;
	}

	static List<List<ImpTerm>> getReverseTab(List<List<ImpTerm>> temp)
	{
		return IntStream.rangeClosed(0,temp.get(0).size()-1)
						.mapToObj(x->
						{
							return IntStream.rangeClosed(0,temp.size()-1)
										    .mapToObj(y->
										    {
										    	return temp.get(y).get(x);
										    }).collect(Collectors.toList());
						}).collect(Collectors.toList());
	}

	private static int[] getFImpIndex(List<List<ImpTerm>> coverTab)
	{
		if(coverTab.isEmpty())return null;
		List<List<List<Integer>>> ti=new ArrayList<>();
		for (int i=0;i<coverTab.size();i++) 
		{
			List<List<Integer>> l=new ArrayList<>();
			for (int k=0;k<coverTab.get(i).size(); k++)if(coverTab.get(i).get(k)!=null)l.add(Arrays.asList(i,k));
			ti.add(l); 
		}
		List<List<List<Integer>>> is=ti.stream().filter(x->x.size()==1).collect(Collectors.toList());
		ti=ti.stream().filter(x->!is.contains(x)).collect(Collectors.toList());
		if(ti.size()==0)return null;
		List<List<Integer>> index=ti.stream().map(x->x.get(0)).collect(Collectors.toList());
		List<List<Integer>> indexs=is.stream().map(x->x.get(0)).collect(Collectors.toList());
		if(index.size()==0&&indexs.size()==1)return new int[]{indexs.get(0).get(0),indexs.get(0).get(1)};		
		List<int[]> g=indexs.stream().filter(x->
		{
			return index.stream().map(y->isEssential(y,x.get(0),x.get(1))).reduce((k,z)->k&&z).get();
		}).map(x->new int[]{x.get(0),x.get(1)}).collect(Collectors.toList());
		if(g.isEmpty())return null;
		return g.get(0);
	}

	public static boolean isEssential(List<Integer> y,int x0,int x1)
	{
		return !((y.contains(x0))||y.contains(x1));
	} 

	public static List<List<ImpTerm>> removeRowAndCulums(int r,List<Integer> culums,List<List<ImpTerm>> coverTab)
	{
		List<List<ImpTerm>> d=new ArrayList<List<ImpTerm>>();
		for(int i=0;i<coverTab.size();i++)
		{
			if(i!=r)
			{
			List<ImpTerm> l=new ArrayList<ImpTerm>();
			for(int k=0;k<coverTab.get(i).size();k++)
				if(!culums.contains(k))l.add(coverTab.get(i).get(k));
			d.add(l);
			}
		}
		return d;
	}

	public static List<List<ImpTerm>> computeCoverTab(List<ImpTerm> fimpTab,List<Term> minTab)
	{
		return fimpTab.stream().map(x->
		{
			return IntStream.rangeClosed(0, minTab.size()-1).mapToObj(y->
			{
				return x.getIndex().contains(minTab.get(y).getIndex())?x:null;	
			}).collect(Collectors.toList());
		}).collect(Collectors.toList());
	}

//Function: getExpVar
//Restituisce la lista delle variabili presenti in exp.

	static List<Character> getExpVar(String exp)
	{
		return exp.chars().filter(x->{return x>=97&&x<=122?true:false;})
				  .distinct()
				  .mapToObj(y->{return Character.valueOf((char)y);})
				  .sorted()
				  .collect(Collectors.toList());
	}

//Function: getCountExpVar
//Restituisce il numero delle variabili presenti in exp.

	static int getCountExpVar(String exp)
	{
		return exp.chars().filter(x->{return x>=97&&x<=122?true:false;})
				  .distinct()
				  .boxed()
				  .collect(Collectors.toCollection(ArrayList::new)).size();
	}

//Function: simplifyTerm
//Restituisce il termine ottenuto applicando l'OR logico a t e t1, tenendo conto delle condizioni di indifferenza "-".

	private static List<String> simplifyTerm(List<String> t,List<String> t1)
	{
		return IntStream.rangeClosed(0, t.size()-1).mapToObj(x->{
			if(t.get(x).equals("-")||t1.get(x).equals("-"))return new String("-");
			return t.get(x).equals(t1.get(x))!=true?new String("-"):new String(t.get(x));
		}).collect(Collectors.toCollection(ArrayList::new));
	}
//Function: getCountTrue
//Restituisce il numero di 1 presenti in term
	private static int getCountTrue(List<String> term)
	{
		return term.stream().filter(x->{return x.equals("1")?true:false;}).collect(Collectors.toCollection(ArrayList::new)).size();
	}
//Function: computeExp
//Calcola e restituisce il valore della funzione in input.
	public static String computeExp(String exp)
	{
		Stack<Boolean> S=new Stack<>();
		for(int i=0;i<exp.length();i++)
		{
			String t= new String()+exp.charAt(i);
			if(t.equals("0"))S.push(false);
			if(t.equals("1"))S.push(true);
			if(t.equals("+"))
			{
				Boolean a=S.pop();
				Boolean b=S.pop();				
				S.push(a||b);
			}
			if(t.equals("*"))
			{
				Boolean a=S.pop();
				Boolean b=S.pop();
				S.push(a&&b);
			}
			if(t.equals("-"))
			{
				Boolean a=S.pop();				
				S.push(!a);
			}
		}
		return S.pop()==true?"1":"0";
	}
	static List<List<String>> getInputValue(int dim)
	{
		return IntStream.rangeClosed(0,(int)Math.pow(2,dim)-1)
				.mapToObj(n->getBinaryRep(n,dim)).collect(Collectors.toCollection(ArrayList::new));
	}
//Function: getBinaryRep
//Restituisce una contenente la rappresentazione binaria di n.	
	private static List<String> getBinaryRep(int n,int dim)
	{
		List<String> l=Stream.iterate(n, x->x/2).limit(dim).map(x->x%2)
				.map(x->x==0?"0":"1").collect(Collectors.toCollection(ArrayList::new));
		List<String> t=new ArrayList<>();
		for(int i=l.size()-1;i>=0;i--)t.add(l.get(i));
			return t;
	}
}
