import java.util.Stack;

//Class: Parser
//Questa classe mette a disposizione i metodi attraverso i quali è possibile stabilire l'appartenenza di una stringa alfanumerica al linguaggio dell'algebra.
//Visto il fine che si propone di realizzare, la classe astrae le "componenti" e i "comportamenti" che caratterizzano gli automi a pila. Formalismi in grado di riconoscere linguaggi di tipo 2.

public class Parser 
{
	//Variable: S
	//Struttura di supporto. Non ha bisogno di presentazioni.
	private Stack<Character> S;
	//Variable: index 
	//L'indice della tabella di parsing.
	private int index;	
	//Variable: Expression
	//L'espressione in input.
	private String Expression;	//	 e   +    *     (    )     $	-
	//Variable: ParsingTable
	//La tabella di parsing.
	private String[][] ParsingTable= {{"TR",null,null,"TR",null,null,"TR"},//E
							 		   {null,"+TR",null,null,"","",null},//R
							 		   {"FP",null,null,"FP",null,null,"FP"},//T
							 		   {null,"","*FP",null,"","",null},//P
							 		   {"e",null,null,"(E)",null,null,"-F"}};//F
	//Constructor: Parser
	public Parser(String Expression) 
	{
		this.Expression=Expression+"$";
		S=new Stack<>();
		S.push(Character.valueOf('$'));
		S.push(Character.valueOf('E'));
		index=0;
		convertVariable();
	}
	//Function: isWellFormed
	//Restituisce true se la l'espressione contenuta in Expression e' ben formata.
	public boolean isWellFormed()
	{
		while(index<Expression.length())
		{
			if(ReadCharacter().equals(S.lastElement()))if(checkIsEnd())return true;
			if(!checkIfExists())return false;	
		}
		return false;
	}
	//Function: checkIsEnd
	//Restituisce true se la pila e' vuota(quindi se l'automa ha riconosciuto la stringa).
	protected boolean checkIsEnd()
	{
		if(S.lastElement().equals(Character.valueOf('$')))return true;
		else 
		{
			S.pop();
			index++;
			return false;
		}
	}

	//Function: checkIfExists
	//Controlla se la tabella di parsing contiene l'elemento ricercato, un mancato riscontro nella ricerca implica la non appartenenza della stringa al linguaggio.  
	
	protected boolean checkIfExists()
	{
		String temp=ParsingTable[getRowIndex(S.lastElement())][getCulumIndex(ReadCharacter())];
		if(temp==null)return false;
		S.pop(); 
		for(int x=temp.length()-1;x>=0;x--)S.push(Character.valueOf(temp.charAt(x)));
		return true;
	}

	//Function: getCulumIndex
	//Codifica il carattere in input nell'indice della tabella tabella di parsing. 

	protected int getCulumIndex(char c)
	{
		switch(c)
		{
			case '-' :
				return 6;
			case '$' :
				return 5;
			case '(' :
				return 3;
			case ')' :
				return 4;
			case '+' :
				return 1;
			case '*' :
				return 2;			
		}
		return 0;
	}

	//Function: getRowIndex
	//Codifica il carattere in input nell'indice della tabella di parsing. 

	protected int getRowIndex(char c)
	{
		switch(c)
		{
			case 'E' :
				return 0;
			case 'R' :
				return 1;
			case 'T' :
				return 2;
			case 'P' :
				return 3;	
			case 'F' :
				return 4;
		}
		return 0;
	}

	//Function: ReadCharacter
	//Legge il prossimo carattere in input.

	protected Character ReadCharacter()
	{
		return Character.valueOf(Expression.charAt(index));		
	}

	//Function: convertVariable
	//Trasforma Expression in una espressione comprensibile all'automa. 

	protected void convertVariable()
	{
		Expression=Expression.replaceAll("[a-z]","e");
	}
}	
