import java.util.Stack;
import java.util.regex.Pattern;

//Class: PostFixCoder
//Questa classe mette a disposizione i metodi necessari codificare l'espressione in notazione infissa, data in input, in una postfissa.
//La classe e' stata pensata come un singleton in quanto non e' necessario all'interno del programma ottenere istanze multiple della suddetta.
public class PostFixCoder 
{
	//Variable: C
	//L'unico oggetto instanziabile della classe. 
	private static PostFixCoder C=null;
	//Variable: S
	private Stack<String> S=null;
	//Variable: postExp 
	private String postExp=null;
	protected PostFixCoder() {}
	//Function: insertSimbol
	private void insertSimbol(String Simbol) {S.push(Simbol);}
	//Function: hasLowerPriority
	//Restituisce true se l'operatore t ha priorita' minore di t1.
	protected boolean hasLowerPriority(String t,String t1)
	{
		if(t.equals("(")||t1.equals(")"))return false;
		return getPriority(t)<getPriority(t1)?true:false;
	}
	//Function: hasSameOrBiggerPriority
	//Restituisce true se l'operatore t ha priorita' maggiore uguale di t1.	
	protected boolean hasSameOrBiggerPriority(String t,String t1)
	{
		if(t.equals("(")||t1.equals(")"))return true;
		return getPriority(t)==getPriority(t1)||getPriority(t)>getPriority(t1)?true:false;
	}
	//Function: getPriority
	//Restituisce la priorita' di t.	
	protected byte getPriority(String t)
	{
		switch(t)
		{
			case "-" : 
				return (byte)3;
			case "*" : 
				return (byte)2;
			case "+" : 
				return (byte)1;			
		}
		return (byte)0;
	}
	//Function: getPostCoder.
	//Questo metodo permette di accedere all'unica istanza di PostFixCoder.
	public static PostFixCoder getPostCoder()
	{
		if(C==null)C=new PostFixCoder();
		return C;
	}
	//Function: getPostFixNotation
	//Restituisce in output expression in notazione postfissa.
	public String getPostFixNotation(String expression)
	{
		S=new Stack<String>();
		postExp=new String();
		for(int i=0;i<expression.length();i++)codElement(expression, i);
		while(!S.empty())postExp+=S.pop();
		return postExp;
	}
	//Function: getPostFixNotation
	//Codifica l'i-esimo elemento di expression.	
	private void codElement(String expression, int i) 
	{
		String t=expression.substring(i,i+1);

		if(Pattern.matches("[a-z]",t))postExp+=t;
		if(Pattern.matches("[+,*,(]",t)||t.equals("-"))codOperator(t);
		if(t.equals(")"))concatStack();
	}
	//Function: concatStack
	//Concatena il contenuto dello stack e lo inserisce in postExp.
	private void concatStack()
	{
		 while(!S.empty()&&!S.lastElement().equals("("))postExp+=S.pop();
		 if(!S.empty())S.pop();
	}
	//Function: codOperator
	//Inserisce nello stack l'operatore p se questo ha una priorita minore uguale.		
	private void codOperator(String t) 
	{
		if(!S.empty())
		{
			if(hasSameOrBiggerPriority(t,S.lastElement()))insertSimbol(t);
			else 
			{
				while(!S.empty()&&hasLowerPriority(t,S.lastElement()))postExp+=S.pop();	
				S.push(t);
			}
		}
		else insertSimbol(t);
	}
	
}