import java.util.ArrayList;
import java.util.List;
import java.util.stream.*;
import java.util.Arrays;
import java.util.regex.*;

// Class: Xbool
//  Contiene tutte le strutture dati utilizzate durante la sintesi e i metodi attraverso i quali viene realizzata.
  
public class Xbool
{
//	Variable: minTab 
//	tabella dei mintermini. Ottenuta da truthTab.	
	private static List<Term> minTab=new ArrayList<>();
//	Variable: truthTab 
//  tabella della verita' della funzione.
	private static List<Term> truthTab=new ArrayList<>();
//	Variable: fimpTab 
//	tabella dei implicanti primi. Ottenuta dal minTab.	
	private static List<ImpTerm> fimpTab=new ArrayList<>();
	private static List<List<ImpTerm>> tempFimpTab=new ArrayList<>();
	private static List<List<ImpTerm>> couplesTermList=new ArrayList<>();  	
//	Variable: eimpTab 
//	tabella contenente tutti gli implicanti primi essenziali (primari e secondari ..) ottenuti a partire dagli implicanti di fimpTab.	
	private static List<ImpTerm> eimpTab=new ArrayList<>();
//	Variable: coverTab 
//	tabella di copertura.
	private static List<List<ImpTerm>> coverTab=new ArrayList<>();
	private static List<List<List<ImpTerm>>> tempCoverTab=new ArrayList<>();
//	Variable: minCoverTab 
//	tabella di copertura.
	private static List<List<ImpTerm>> minCoverTab=new ArrayList<>();
//  Variable: variables 
//	lista contenente le variabili utilizzate nell'espressione.	   
	private static List<Character> variables=new ArrayList<>(); 
	private static boolean circular=false;
	public static void main(String[] args) 
	{
		if(args.length<1||args[0].equals("-h")){printHelp();return;}
		for (int i=0;i<args[0].length();i++) 
			if(!hasAdmittedCharacter(args[0].substring(i,i+1))){System.out.println("Error! Wrong expression's syntax !"); return;}
		if(!new Parser(args[0]).isWellFormed()) {System.out.println("Error! Wrong expression's syntax !"); return;}
		if(args.length==2&&args[1].equals("-v"))quineMcCluskeyVerbose(args[0]);
		else quineMcCluskey(args[0]);
	}
	private static boolean hasAdmittedCharacter(String t)
	{
		return Pattern.matches("[a-z]",t)||Pattern.matches("[+,*,(,),-]",t)?true:false;
	}
//  Function: printHelp
//	    Stampa a terminale le istruzioni d'uso del programma.	
	private static void printHelp()
	{
		System.out.print("Xbool 1.0 - a tool for boolean algebra\n\n"
				+ "author Gianluca Trigili\n\n"
				+ "This tool can find a minimum synthesis boolean function by any boolean function.\n"
				+ "Syntax example for minimum synthesis boolean function: \njava Xbool <expression> \n"
				+ "Expression can be any well formed infix boolean algebraic expression with almost 26 variables.\n"
				+ "Only the lower latin letter are admited.\n"
				+ "The boolean's operator are coded as follow "
				+ "AND = * , OR = + , NOT = - .\n"
				+ "The NOT operator has the highest priority, followed by AND, followed by OR.\n"
				+ "Each expression where parentheses are omitted by associativity of the binary operators,\n"
				+ "is considered as left associated expression : a+b+c -> ((a+b)+c).\n"
				+ "Example of expression : (a+b)*-c .\n"
				+ "The following expressions are interpreted in the same way : \n"
				+ "(a+b)*c,(a+b)*(c),((a+b)*c),\n"
				+ "which means that the delimiters are considered only if they are important.\n"
				+ "Example : java Xbool \"(a+b)*c\"\n"
				+ "If you want to see more details about the intermediate steps \n"
				+ "of the Quine-McCluskey process you can add the -v flag.\n"
				+ "With the -v flag the program output shows the coverage tables that are generated during the synthesis process.\n"
   				+ "Each row, of a coverage table relates to a prime implicant previously identified during the expansion process.\n"
   				+ "Each column, on the other hand, relates to a minterm that occurs in the SOP form of the same functions.\n"
   				+ "The row with index i refers to the prime implicant with the same index in the table of prime implicants.\n" 
   				+ "The column with index j refers to the minterm with the same index in the minterms table.\n");
	}
//   Function: computeMinSinFunc
//  	Richiama tutti i metodi necessari al compimento dell'algoritmo

	private static String computeMinSinFunc(String t)
	{
		variables=ExpSolver.getExpVar(t);
		truthTab=ExpSolver.computeTruthTable(ExpSolver.getInputValue(ExpSolver.getCountExpVar(t)),variables,t);
		minTab=ExpSolver.computeMinTermTab(truthTab);
		fimpTab=ExpSolver.computeFimpTab(t,minTab,fimpTab,tempFimpTab,couplesTermList);
		if(fimpTab.isEmpty())return getExpRep(minTab.stream().map(x->{return new ImpTerm(x);}).collect(Collectors.toList()));
		coverTab=ExpSolver.computeCoverTab(fimpTab,minTab);
		minCoverTab=ExpSolver.computeMinCover(coverTab,eimpTab,tempCoverTab);
		tempCoverTab=tempCoverTab.stream().distinct().collect(Collectors.toList());
		if(minCoverTab.isEmpty())return getExpRep(eimpTab);
		circular=true;
		List<ImpTerm> fsolution=new ArrayList<>();
		return getExpRep(branchAndBound(fsolution,minCoverTab));
	}
	private static List<ImpTerm> branchAndBound(List<ImpTerm> fsolution,List<List<ImpTerm>> coverTab)
	{
		List<List<ImpTerm>> minCoversF=new ArrayList<>();
		for (int i=0;i<coverTab.size();i++) 
		{
			List<List<ImpTerm>> tempCt=coverTab.stream().collect(Collectors.toList()),_tempCt=null;
			List<ImpTerm> solution=new ArrayList<>();
			solution.add(getFimp(coverTab.get(i)));
			tempCt=ExpSolver.removeRowAndCulums(i,ExpSolver.getMinTermsIndex(coverTab.get(i)),coverTab);
			_tempCt=ExpSolver.computeMinCover(tempCt,solution,new ArrayList<>());			
			if(_tempCt.isEmpty())minCoversF.add(Stream.concat(solution.stream(),fsolution.stream()).collect(Collectors.toList()));
			else minCoversF.add(Stream.concat(fsolution.stream(),branchAndBound(solution,_tempCt).stream()).collect(Collectors.toList()));
		}
		return minCoversF.stream().min((x,y)->x.size()-y.size()).get();
	}	
	private static ImpTerm getFimp(List<ImpTerm> l)
	{
		for (int i=0;i<l.size();i++)if(l.get(i)!=null)return l.get(i);
		return null;
	}
	private static void quineMcCluskeyVerbose(String exp) 
	{
		String t=PostFixCoder.getPostCoder().getPostFixNotation(exp);
		t=computeMinSinFunc(t);
		System.out.println("Input function = "+exp+"\n");
		System.out.println("Table of truth\n");
		printTermsTab(truthTab);
		System.out.println("\nTable of minterms\n");
		printTermsTab(minTab);
		System.out.println("\nTables of implicants\n");
		tempFimpTab.stream().forEach(x->{printImpTab(x);System.out.println();});
		System.out.println("\nExpanded implicants\n");
		couplesTermList.stream().forEach(x->System.out.println(x.get(0).getTerms()+" "+x.get(1).getTerms()+" -> "+x.get(2).getTerms()));
		System.out.println("\nTables of coverage\n");
		tempCoverTab.stream().forEach(x->printCoverTab(x));
		if(circular)System.out.println("\nCircular table of prime implicants detected. Branch and bound algorithm applied.");
		System.out.println("\nMinimum syntax function: "+t);
	}
	private static void quineMcCluskey(String exp) 
	{
		String t=PostFixCoder.getPostCoder().getPostFixNotation(exp);
		t=computeMinSinFunc(t);
		System.out.println("Input function = "+exp+"\n");
		System.out.println("Table of truth\n");
		printTermsTab(truthTab);
		if(circular)System.out.println("\nCircular table of prime implicants detected. Branch and bound algorithm applied.");		
		System.out.println("\nMinimum syntax function: "+t);
	}
//  Function: getExpRep    	
//    	Restituisce la rapressentazione alfanumerica della funzione minimizzata.	
	private static String getExpRep(List<ImpTerm> terms)
	{
		return terms.stream().filter(x->x!=null).map(x->
		{
			return getTermRep(x);
		}).collect(Collectors.joining("+"));
	}
//  Function: getTermRep
//		Restituisce la reppresentazione dell'implicante primo in input per la tabella di copertura.
	private static String getTermRep(ImpTerm x)
	{
		return IntStream.rangeClosed(0, x.getTerms().size()-1).mapToObj(y->
		{
			return x.getTerms().get(y).equals("-")?",":getStringRep(x.getTerms().get(y),y);
		}).filter(w->!w.equals(",")).collect(Collectors.joining("*"));
	}
//	Function: getStringRep
//		Restituisce il letterale associato all'elemento in input. 	
	private static String getStringRep(String element,int index)
	{
		return element.equals("0")?"-"+variables.get(index):""+variables.get(index);
	}
	private static void printTermsTab(List<Term> tab)
	{
		System.out.println(variables.stream().map(x->{return new String()+x;}).collect(Collectors.joining(" | ")));
		if(tab!=null)tab.stream().forEach(x->
		{
			System.out.println(x.getTerms().stream().collect(Collectors.joining(" | "))+"\t"+x.getValue()+"\t["+x.getIndex()+"]");
		});		
	}
	private static void printImpTab(List<ImpTerm> tab)
	{
		if(tab!=null)tab.stream().forEach(x->
		{
			System.out.println(x.getTerms()+"\t"+x.getIndex());
		});		
	}
	private static void printCoverTab(List<List<ImpTerm>> tab)
	{
		if(tab!=null)tab.stream().forEach(x->
		{
			System.out.println("|"+x.stream().map(y->
			{
				return y!=null?"x":"-";
			}).collect(Collectors.joining("|"))+"|");
		});
		System.out.println();
	}
}
